<html>
 <head><title>Selected Chord</title></head>
 <body>
<?
   require('scales.inc');

   define("FRET_OFFSET", "fret_offset");
   define("NOTE",        "note");
   define("FRET_NONE",   -1);
   define("NOTE_NONE",   "none");
   define("NCHORD_FRETS",     4);
   define("MIN_NCONSECUTIVE", 4);
   define("NSEARCH_FRETS",   12);
   define("SCORE_FACTOR",    10);

   class ChordConfigurationSupplier
   {
      var $nConfigurations    = false;
      var $configurationCount = false;
      var $values             = false;
      var $indices            = false;
      var $counts             = false;

      function ChordConfigurationSupplier(
         $notesInChord,
         $displayMode,
         $fretNumber
      )
      {
         $this->values  = array();
         $this->indices = array();
         $this->counts  = array();
         $nChordNotes  = count($notesInChord);
         $openStrings  = FretBoard::getOpenStrings();
         $notes        = Notes::get($displayMode);
         $nNotes       = count($notes);

         $this->nConfigurations    = 1;
         $this->configurationCount = 0;

         for($i = 0; $i < FRETBOARD_NSTRINGS; $i ++)
         {
            $this->values[$i] = array(
               array(
                  FRET_OFFSET => FRET_NONE,
                  NOTE        => NOTE_NONE
               )
            );

            $offset = Notes::getOffset(
               $openStrings[$i], $displayMode
            );

            for($j = 0; $j < NCHORD_FRETS; $j ++)
            {
               $thisNote = $notes[($offset + $fretNumber + $j) % $nNotes];

               // Is this note in '$notesInChord'?

               $found = false;

               for($k = 0; $k < $nChordNotes; $k ++)
               {
                  if($thisNote == $notesInChord[$k])
                  {
                     $found = true;
                     break;
                  }
               }

               if($found)
               {
                  $this->values[$i][] = array(
                     FRET_OFFSET => $j,
                     NOTE        => $thisNote
                  );
               }
            }

            $size = count($this->values[$i]);

            $this->nConfigurations *= $size;
            $this->counts[$i]       = $size;
            $this->indices[$i]      = 0;
         }
      }

      function hasMoreConfigurations()
      {
         return($this->configurationCount < $this->nConfigurations);
      }

      function nextConfiguration()
      {
         $config = false;

         if($this->hasMoreConfigurations())
         {
            $config = array();

            for($i = 0; $i < FRETBOARD_NSTRINGS; $i ++)
            {
               $config[$i] = $this->values[$i][$this->indices[$i]];
            }

            $this->configurationCount++;

            if($this->hasMoreConfigurations())
            {
               $i = 0;

               while(true)
               {
                  if($this->indices[$i] + 1 < $this->counts[$i])
                  {
                     $this->indices[$i]++;
                     break;
                  }

                  $this->indices[$i] = 0;
                  $i++;
               }
            }
         }

         return($config);
      }
   }

   function printConfig($config)
   {
      print("<pre>\n");

      for($i = FRETBOARD_NSTRINGS - 1; $i >= 0; $i --)
      {
         $fret = $config[$i][FRET_OFFSET];
         $note = sprintf("%2s", $config[$i][NOTE]);
   
         print("[" . ($fret == 0 ? "$note" : "  ") . "]");
         print("[" . ($fret == 1 ? "$note" : "  ") . "]");
         print("[" . ($fret == 2 ? "$note" : "  ") . "]");
         print("[" . ($fret == 3 ? "$note" : "  ") . "]");
         print("\n");
      }
   
      print("\n");
      print("</PRE>\n");
   }

   class ChordData extends FretBoardData
   {
      var $table;
      var $title;

      function ChordData($title, $table)
      {
         $this->title = $title;
         $this->table = $table;
      }

      function getTitle() { return($this->title); }

      function displayNote($theNote, $stringNumber, $fretNumber)
      {
         return(
            isset($this->table[$fretNumber . ":" . $stringNumber])
         );
      }
   }

   function ChordCallback(&$clientData, $thisNote)
   {
      $clientData[] = $thisNote;
   }

   class ChordTable
   {
      var $score;
      var $table;
      var $fretNumber;

      function ChordTable($fretNumber, $config, $score)
      {
         $this->table      = array();
         $this->score      = $score;
         $this->fretNumber = $fretNumber;

         for($i = 0; $i < FRETBOARD_NSTRINGS; $i ++)
         {
            $note = $config[$i][NOTE];
            $fret = $config[$i][FRET_OFFSET];
   
            if($fret != FRET_NONE)
            {
               $adjustedFret = $fret + $fretNumber;
               $this->table[$adjustedFret . ":" . $i] = true;
            }
         }
      }

      function getTable() { return($this->table); }
      function getScore() { return($this->score); }

      function isSubsetOf($another)
      {
         // If all of our $table entries are found within
         // the table of $another we're a subset

         $isSubset   = true;
         $otherTable = $another->getTable();

         reset($this->table);

         while(list($key, $value) = each($this->table))
         {
            if(!isset($otherTable[$key]))
            {
               $isSubset = false;
               break;
            }
         }

         reset($this->table);

         return($isSubset);
      }

      function display()
      {
         $values = array();

         for($i = 0; $i < FRETBOARD_NSTRINGS; $i ++)
         {
            $entries = array();

            for($j = 0; $j < NCHORD_FRETS; $j ++)
               $entries[$j] = " ";

            $values[] = $entries;
         }

         reset($this->table);

         while(list($key, $value) = each($this->table))
         {
            $numbers = explode(":", $key);
            $fret    = $numbers[0] - $this->fretNumber;
            $string  = $numbers[1];

            $values[$string][$fret] = "X"; 
         }

         reset($this->table);

         print("<pre>\n");

         for($i = FRETBOARD_NSTRINGS - 1; $i >= 0; $i --)
         {
            for($j = 0; $j < NCHORD_FRETS; $j ++)
               print("[" . $values[$i][$j] . "]");

            print("\n");
         }

         print("\n");
         print("</PRE>\n");
      }
   }

   function sortTableEntry($first, $second)
   {
      $firstScore  = $first->getScore();
      $secondScore = $second->getScore();

      if($firstScore > $secondScore)
         return(-1);

      if($firstScore < $secondScore)
         return(1);

      return(0);
   }

   /*****************************************************************
    * Script Begins Here
    *****************************************************************/

   $chordNotes    = array();
   $chords        = Chords::get();
   $key           = Scales::getRequestValue("key");
   $chordName     = Scales::getRequestValue("chordName");
   $displayMode   = Scales::getRequestValue("displayMode");
   $chordPattern  = $chords[$chordName];
   $title         = "";

   PatternHandler::process(
      $chordNotes,
      $title,
      $key,
      $chordPattern,
      $displayMode,
      "ChordCallback"
   );

   $title = "$key $chordName " . $title;

   $data = new ScaleFretBoardData(
      $key, $chordName, $chordPattern, $displayMode, "All notes in: "
   );

   FretBoard::draw($data, $displayMode);

   print("<p>\n");

   $nChordNotes = count($chordNotes);

   for($fretNumber = 0; $fretNumber < NSEARCH_FRETS; $fretNumber ++)
   {
      $supplier = new ChordConfigurationSupplier(
         $chordNotes, $displayMode, $fretNumber
      );

      $bestScore = 0;
      $tables    = array();

      while($supplier->hasMoreConfigurations())
      {
         $config = $supplier->nextConfiguration();

         //printConfig($config);
   
         // Ok, many configurations are immediately discardable so
         // begin the weedout process.  First discard any configuration
         // that has no notes in fret offset 0 ...
   
         $notesInFretZero = 0;
   
         for($i = 0; $i < FRETBOARD_NSTRINGS; $i ++)
         {
            $fret = $config[$i][FRET_OFFSET];
   
            if($fret == 0)
            {
               $notesInFretZero++;
               break;
            }
         }
   
         if($notesInFretZero == 0)
         {
            //print("Discarded, no notes in fret zero<br>\n");
            continue;
         }
   
         // Now we discard any configuration that has less than
         // MIN_NCONSECUTIVE consecutive strings with a note in the
         // chord
   
         $nConsecutive = 0;
         $count        = 0;
   
         for($i = 0; $i < FRETBOARD_NSTRINGS; $i ++)
         {
            $fret = $config[$i][FRET_OFFSET];
   
            if($fret != FRET_NONE)
            {
               $count++;
               $nConsecutive = max($count, $nConsecutive);
            }
            else
               $count = 0;
         }
   
         if($nConsecutive < MIN_NCONSECUTIVE)
         {
            //print("Discarded, two few consecutive strings<br>\n");
            continue;
         }
   
         // Now we wish to discard any configuration that doesn't have all
         // the notes in the chord
   
         $counts = array();
   
         for($i = 0; $i < $nChordNotes; $i ++)
            $counts[$chordNotes[$i]] = 0;
   
         for($i = 0; $i < FRETBOARD_NSTRINGS; $i ++)
            if(($note = $config[$i][NOTE]) != NOTE_NONE)
               $counts[$note]++;
   
         $foundAllNotes = true;
   
         while(list($theNote, $theCount) = each($counts))
         {
            if($theCount == 0)
            {
               $foundAllNotes = false;
               break;
            }
         }
   
         if(!$foundAllNotes)
         {
            //print("Discarded, not all notes found<br>\n");
            continue;
         }
   
         // Now we weed out "unplayable" configurations, for now "unplayable"
         // means: "More than 4 notes beyond fret offset 0 or more than
         // 3 notes beyond fret offset zero if we're not in the first
         // position" though we might want to smarten this up in time
   
         $noteCount = 0;
   
         for($i = 0; $i < FRETBOARD_NSTRINGS; $i ++)
         {
            $fret = $config[$i][FRET_OFFSET];
   
            if($fret != FRET_NONE && $fret > 0)
               $noteCount++;
         }
   
         if($noteCount > 4 || ($noteCount > 3 && $fretNumber > 0))
         {
            //print("Discarded, unplayable<br>\n");
            continue;
         }

         $additionalScore = 0;
   
         for($i = 0; $i < FRETBOARD_NSTRINGS; $i ++)
         {
            if(($note = $config[$i][NOTE]) == $chordNotes[0])
               $additionalScore += (FRETBOARD_NSTRINGS  - $i);
         }
   
         $score = $nConsecutive * SCORE_FACTOR + $additionalScore;
         $entry = new ChordTable($fretNumber, $config, $score);

         $subsetFlag = false;
         $size       = count($tables);

         for($i = 0; $i < $size; $i ++)
         {
            if($entry->isSubsetOf($tables[$i]))
            {
               $subsetFlag = true;
               break;
            }
         }

         if(!$subsetFlag)
         {
            $newTables = array();

            for($i = 0; $i < $size; $i ++)
            {
               $thisEntry = $tables[$i];

               if(!$thisEntry->isSubsetOf($entry))
                  $newTables[] = $thisEntry;
            }

            $newTables[] = $entry;
            $tables      = $newTables;
         }
      }

      if(($size = count($tables)) > 0)
      {
         print("<hr>\n");

         // sort the entries by score ...

         usort($tables, "sortTableEntry");

         for($i = 0; $i < $size; $i ++)
         {
            $data = new ChordData($title, $tables[$i]->getTable());
            FretBoard::draw($data, $displayMode);

            print("<p>\n");
         }
      }
   }
?>
   <p>
   Return to <a href="choose_chord.php">choose chord</A>.
   </p>
  </body>
</html>
